/*
 *   Nanova: library for unbalanced nestead ANOVA
 *   
 *   Copyright (C) 2006 Claire Bardel, Vincent Danjean
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "nanova.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_cdf.h>

#include <clapack.h>

#define erreur(str, ...) {\
        fprintf(stderr, "nanova: " str, ##__VA_ARGS__);\
}

#define DEBUG 0

struct grp_level {
	unsigned int nval;
	int parent;
	double sumval;
	double moyenne;
};

static int fill_groups_per_level(unsigned int nb_levels, 
				 unsigned int nb_groups,
				 unsigned int nb_values, 
				 unsigned int tree[nb_groups][nb_levels],
				 int groups_per_level[nb_levels+2])
{
	int i;
	groups_per_level[0]=1;
	groups_per_level[nb_levels+1]=nb_values;
	for (i=1; i<=nb_levels; i++) {
		groups_per_level[i]=tree[nb_groups-1][i-1]+1;
	}
		
#if DEBUG 	
	printf("groups: ");
	for (i=0; i<=nb_levels+1; i++) {
		printf ("%d ", groups_per_level[i]);
	}

#endif

	return 0;
}


static int fill_nval(unsigned int nb_levels, unsigned int nb_groups,
		     unsigned int tree[nb_groups][nb_levels],
		     unsigned int groups[nb_groups],
		     struct grp_level data[nb_groups][nb_levels+1])
{
	int i,j;
	for(i=0; i<nb_groups; i++) {
		int effectif=groups[i];
		data[0][0].nval += effectif;
		for(j=1; j<=nb_levels; j++) {
			int group=tree[i][j-1];
			if (group >= nb_groups) {
				erreur("in tree: for groups of last-level"
				       " group %i, the group of level %i is %i"
				       " and there should be only %i groups"
				       " (0 to %i)\n",
				       i, j, tree[i][j-1], nb_groups,
				       nb_groups-1);
				return 1;
			}
			data[group][j].nval += effectif;
		}
	}

	for(j=1; j<=nb_levels; j++) {
		int total=0;
		for(i=0; i<nb_groups; i++) {
			if (data[i][j].nval==0) {
				if (total!=data[0][0].nval) {
					erreur("in tree: in level %i, group %i"
					       " is empty (groups must be"
					       " continuous)\n", j, i);
					return 1;
				}
				break;
			}
			total+=data[i][j].nval;
		}
	}
	
	return 0;
}

static int fill_parent(unsigned int nb_levels, unsigned int nb_groups,
		       unsigned int tree[nb_groups][nb_levels],
		       struct grp_level data[nb_groups][nb_levels+1])
{
	int i,j;

	for(i=0; i<nb_groups; i++) {
		for(j=0; j<=nb_levels; j++) {
			data[i][j].parent=-1;
		}
	}
	for(i=0; i<nb_groups; i++) {
		int father=0;
		for(j=1; j<=nb_levels; j++) {
			int group=tree[i][j-1];
			if (data[group][j].parent != -1 &&
			    data[group][j].parent != father) {
				erreur("in tree: group %i of level %i has two"
				       " parents (%i and %i)\n",
				       group, j, data[group][j].parent,
				       father);
				return 1;
			}
			//printf("(%d, %d) ", tree[i][j-1], j);
			data[group][j].parent = father;
			father=group;
		}
		//printf("\n");
	}
	
	return 0;
}


static int fill_gamma (unsigned int nb_levels, unsigned int nb_groups,
		       unsigned int tree[nb_groups][nb_levels],
		       struct grp_level data[nb_groups][nb_levels+1],
		       int groups_per_level[nb_levels+2],
		       double gamma[nb_levels+1][nb_levels+1]
	)
{
	int i, j, k, l;
	memset(gamma, 0, (nb_levels+1)*(nb_levels+1)*sizeof(double));

	for (j=nb_levels; j>=0; j--) {
		//printf("LEVEL %d\n", j);
		for (k=0; k<groups_per_level[j]; k++) { 
			//printf("  k= %d\n", k);	
			int num=(data[k][j].nval)*(data[k][j].nval);
			l=k;
			//printf("  num= %d\n", num);
			gamma[j][j]=data[0][0].nval;
			for (i=j-1; i>=0; i--) {
				//printf("    i= %d ",i);
					
				l=data[l][i+1].parent;
				int denom=data[l][i].nval;
			       	//printf("denom= %d num= %d", denom, num);
			       	//double div=(double)num/ (double)denom;
			       	//printf(" div= %g ", div);
				gamma[i][j]+=(double) num/(double) denom;
			       	//	printf("gamma= %g \n", gamma[i][j]);
			}
			
		}
	}
					
#if DEBUG
	for(i=0; i<=nb_levels; i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%.3f, ", gamma[i][j]);
		}
		printf("\n");
	}
#endif

	return 0;
}

static int fill_ddl(unsigned int nb_levels, 
	     int groups_per_level[nb_levels+2], 
	     int ddl[nb_levels+1]) 
{
	int i;
	memset(ddl, 0, (nb_levels+1)*sizeof(int));

	for (i=0; i<=nb_levels; i++) {
		ddl[i]=groups_per_level[i+1]-groups_per_level[i];
	}

#if DEBUG
	printf("ddl: ");
	for (i=0; i<=nb_levels; i++) {
		printf("%d ", ddl[i]);
	}
	printf("\n");
#endif

	return 0;
}


static int fill_c(unsigned int nb_levels,  
	   int ddl[nb_levels+1], 
	   double gamma[nb_levels+1][nb_levels+1],
	   double c[nb_levels+1][nb_levels+1])
{
	int i, j;
	memset(c, 0, (nb_levels+1)*(nb_levels+1)*sizeof(double));
	
	for (i=0; i<=nb_levels; i++) {
		for (j=i; j<=nb_levels; j++) {
			if (j == nb_levels) {
				c[i][j]=1;
			} else {
				c[i][j]=(gamma[i+1][j+1]-gamma[i][j+1])/
					(double)(ddl[i]);
			}
		}
	}
#if DEBUG
	printf("matrice c\n");
	for(i=0; i<=nb_levels; i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%.3f, ", c[i][j]);
		}
		printf("\n");
	}
#endif
	return 0;
}

static int fill_sumval(unsigned int nb_levels, unsigned int nb_groups, 
		unsigned int nb_values, 
		struct grp_level data[nb_groups][nb_levels+1], 
		double values[nb_values],
		int group_per_level[nb_levels+2])
{
	int i, j;
// Fill sum val for the last level
	int decale = 0;
	for (i=0; i<nb_groups; i++) {
		int nb_values_group=data[i][nb_levels].nval;
		for (j=0; j<nb_values_group; j++) {
			data[i][nb_levels].sumval+=values[decale+j];
		}
		decale+=nb_values_group;
	}
			
// Fill sumval for the other levels
	for (j=nb_levels; j>0; j--) {
		for (i=0; i<group_per_level[j]; i++) {
			int father =data[i][j].parent;
			data[father][j-1].sumval+=data[i][j].sumval;
		}
	}

#if DEBUG
	printf("matrice sumval\n");
	for(i=0; i<nb_groups; i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%.3f, ", data[i][j].sumval);
		}
		printf("\n");
	}
#endif

	return 0;
}

static int fill_moyennes (unsigned int nb_levels, unsigned int nb_groups, 
		   struct grp_level data[nb_groups][nb_levels+1],
		   int group_per_level[nb_levels+2])
{
	int i, j;
	for (j=nb_levels; j>=0; j--) {
		for (i=0; i<group_per_level[j]; i++) {
			data[i][j].moyenne=data[i][j].sumval/data[i][j].nval;
		}
	}	
#if DEBUG
	printf("matrice moyennes\n");
	for (i=0; i<nb_groups; i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%.3f, ", data[i][j].moyenne);
		}
		printf("\n");
	}
#endif
	return 0;
}

/* VINCE Est-ce que c'est plus judicieux de calculer directement les
 * carrs moyens plutot que les sommes des carrs des carts?*/

static int fill_ss(unsigned int nb_levels, unsigned int nb_groups, 
	     unsigned int nb_values, 
	     struct grp_level data[nb_groups][nb_levels+1], 
	     double values[nb_values],
	     int group_per_level[nb_levels+2], double ss[nb_levels+1])
{
	int i, j;

	memset(ss, 0, (nb_levels+1)*sizeof(double));
	// Cas du dernier level: somme des carrs intra
	int decale = 0;
	for (i=0; i<nb_groups; i++){
		int nb_value_grp=data[i][nb_levels].nval;
		for (j=0;j<nb_value_grp; j++) {
			ss[nb_levels]+=(values[decale+j]-data[i][nb_levels].moyenne)
				*(values[decale+j]-data[i][nb_levels].moyenne);
//printf("sse n = (%f-%f)^2\n", values[decale+j], data[i][nb_levels].moyenne);
		}
//printf("decal = %d\n", decale);
		decale+=nb_value_grp;
	}
//printf("total=%f\n",ss[nb_levels]); 
	
        // Traitement des autres niveaux (sommes inter sous-groupes)
	for (j=nb_levels; j>0; j--) {
		for (i=0; i<group_per_level[j]; i++) {
			int father = data[i][j].parent;
			ss[j-1]+=(data[i][j].moyenne-data[father][j-1].moyenne)
				*(data[i][j].moyenne-data[father][j-1].moyenne)
				*data[i][j].nval;
			//	printf("sse %d = (%f-%f)^2*%d\n", j-1, data[i][j].moyenne, data[father][j-1].moyenne, data[i][j].nval); 
		}
		// printf("total=%f\n",ss[j-1]); 
	}
	
	
#if DEBUG
	printf("SS:");
	for (i=0; i<=nb_levels; i++) {
		printf("\t[%d]=%f", i, ss[i]);
	}
	printf("\n");
#endif
	return 0;
}

/* Pour vrifier si la somme des carrs totale est gale  la somme des 
sommes des carrs*/
static int verif_sst (unsigned int nb_levels, unsigned int nb_groups,
	       unsigned int nb_values, 
	       double ss[nb_levels+1],
	       struct grp_level data[nb_groups][nb_levels+1],
	       double values[nb_values])
{
	int i;
	double sst=0.0, sst_calc=0.0;
	double moyenne=data[0][0].moyenne;
	for (i=0; i<nb_values; i++) {
	  sst += (values[i]-moyenne)*(values[i]-moyenne);
	}
	
	for (i=0; i<=nb_levels; i++) {
	  sst_calc += ss[i];
	}

	double pourcent = (sst - sst_calc)/sst_calc;
	if (fabs(pourcent) > 0.01) {
		erreur ("the total sum of square %f is different from the"
			" sum of the sum of squares, %f \n", sst, sst_calc);
		return 1;
	} else {
#if DEBUG
		printf("sum of square = %f\nsum of the sum of squares=%f\n",
			sst, sst_calc);
#endif
		return 0;
	}
}


static int fill_ms (unsigned int nb_levels, double ss[nb_levels+1], 
	     int ddl[nb_levels+1], double ms[nb_levels+1] )
{
	int i;
	for (i=0; i<=nb_levels; i++) {
		ms[i]=ss[i]/ (double)ddl[i];
	}

#if DEBUG
	printf("MS:");
	for (i=0; i<=nb_levels; i++) {
		printf("\t[%d]=%f", i, ms[i]);
	}
	printf("\n");
#endif
	return 0;	
}

static int fill_F (unsigned int nb_levels, double denom[nb_levels], 
	    double ms[nb_levels+1], struct una_results res[nb_levels])
{
	int i;
	for (i=0; i<nb_levels; i++) {
		res[i].F=ms[i]/denom[i];
	}
#if DEBUG
	printf("F:");
	for (i=0; i<nb_levels; i++) {
		printf("\t[%d]=%f", i, res[i].F);
	}
	printf("\n");
#endif
	return 0;	
	
}

int unbalanced_nested_anova(unsigned int nb_levels, unsigned int nb_groups,
			    unsigned int nb_values, 
			    unsigned int *tree, unsigned int *groups,
			    double *values,
			    struct una_results *res)
{
	int i, j;// k, l, m;
	
	struct grp_level data[nb_groups][nb_levels+1];

	double gamma[nb_levels+1][nb_levels+1];
	int groups_per_level[nb_levels+2];
	int ddl[nb_levels+1];
	double c[nb_levels+1][nb_levels+1];
	double ss[nb_levels+1]; //Attention, dans case 0, c'est sse level 1
	                         //Le dernier level correspond  l'intra
	double ms[nb_levels+1]; // Meme remarque que pour ss
	memset(data, 0, (nb_levels+1)*nb_groups*sizeof(struct grp_level));

#define CHECK_RET(f, args) \
    { \
        int ret=f args; \
        if (ret) { \
            erreur("error while invoking " #f " (return code: %d)\n", ret); \
            return ret; \
        } \
    }
	
	CHECK_RET(fill_groups_per_level,(nb_levels, nb_groups, nb_values, 
					(void*)tree, groups_per_level));
	CHECK_RET(fill_nval,(nb_levels, nb_groups, (void*)tree, groups, data));
	CHECK_RET(fill_parent,(nb_levels, nb_groups, (void*)tree, data));
	CHECK_RET(fill_gamma,(nb_levels, nb_groups, (void*)tree, data,
			     groups_per_level, gamma));
	CHECK_RET(fill_ddl,(nb_levels, groups_per_level, ddl));
	CHECK_RET(fill_c,(nb_levels, ddl, gamma, c));
	CHECK_RET(fill_sumval,(nb_levels, nb_groups, nb_values, data, values,
			      groups_per_level));
	CHECK_RET(fill_moyennes,(nb_levels, nb_groups, data, groups_per_level));
	CHECK_RET(fill_ss,(nb_levels, nb_groups, nb_values, data, values,
			  groups_per_level, ss));
	CHECK_RET(fill_ms,(nb_levels, ss, ddl, ms));
	CHECK_RET(verif_sst,(nb_levels, nb_groups, nb_values, ss, data, 
			     values));

#if 0
	printf("***************************************\n"
	       "* SWITCHING DATA (c, ms and ddl) !!!! *\n"
	       "***************************************\n");
	if (nb_levels!=3) {
		erreur("Bad nb_levels\n");
		return 1;
	}
	double my_c[]= {
		 7, 3, 1.571, 1 ,
		 0, 2, 1.214, 1 ,
		 0, 0, 1.5,   1 ,
		 0, 0, 0,     1 ,
	};
	double my_ms[]= {
		0.18404, 0.04508, 0.01625, 0.01004
	};
	int my_ddl[]={10, 22, 22, 22, 50};

	printf("\nddl:");
	for (i=0; i<= nb_levels; i++) {
		for (j=0; j<=nb_levels; j++) {
			c[i][j]=my_c[i*4+j];
		}
		ms[i]=my_ms[i];
		ddl[i]=my_ddl[i];
		printf("\t[%i]=%i", i, ddl[i]);
	}
#endif

	double cinv[nb_levels+1][nb_levels+1];
	memcpy(cinv, c, sizeof(c));
	CHECK_RET(clapack_dtrtri,(CblasRowMajor, CblasUpper, CblasNonUnit,
				  nb_levels+1, (double*)cinv, nb_levels+1));

#if DEBUG
	printf("\ncinv:\n");
	for(i=0; i<=nb_levels;i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%g, ", cinv[i][j]);
		}
		printf("\n");
	}
#endif
	
	double sigma_hat[nb_levels+1];
	memcpy(sigma_hat, ms, sizeof(ms));
	int ipiv[nb_levels+1];
	CHECK_RET(clapack_dgesv,(CblasRowMajor, nb_levels+1, 1,
				 (double*)c, nb_levels+1, (int*)ipiv,
				 (double*)sigma_hat, nb_levels+1));

	double denom[nb_levels];
	double ddl_denom[nb_levels];

	for (i=0; i<nb_levels; i++) {
		double sum=0;
		for(j=i+1; j<=nb_levels; j++) {
			double d=c[i][i]*cinv[i][j]*ms[j];
			sum+=d*d/ddl[j];
		}
		denom[i]=ms[i]-c[i][i]*sigma_hat[i];
		ddl_denom[i]=denom[i]*denom[i]/sum;
		res[i].F=ms[i]/denom[i];
	}

	for (i=0; i<nb_levels; i++) {
		res[i].p_value=gsl_cdf_fdist_Q(res[i].F, (double)ddl[i], 
					       ddl_denom[i]);
	}   
	
//double gsl_cdf_fdist_Q (double x, double nu1, double nu2)
#if DEBUG
	printf("\nnval:\n");
	for(i=0; i<nb_groups; i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%2d, ", data[i][j].nval);
		}
		printf("\n");
	}

	printf("c=%p, cinv=%p (size : %d=%d^2*%d)\n", c, cinv, sizeof(cinv),
	       nb_levels+1, sizeof(double));

	printf("\nc:\n");
	for(i=0; i<=nb_levels;i++) {
		for(j=0; j<=nb_levels; j++) {
			printf("%g, ", c[i][j]);
		}
		printf("\n");
	}
	printf("\nsigma_hat:");
	for(i=0; i<=nb_levels;i++) {
		printf("\t[%i]=%g", ipiv[i], sigma_hat[i]);
	}
	printf("\ndenom:");
	for(i=0; i<nb_levels;i++) {
		printf("\t[%i]=%g", i, denom[i]);
	}
	printf("\nddl_denom:");
	for(i=0; i<nb_levels;i++) {
		printf("\t[%i]=%g", i, ddl_denom[i]);
	}
	printf("\n");
#endif
	return 0;
}

